"""Handle Git-related functionality."""
import string

import pygit2


def get_git_status(repo_path="."):
    """
    Retrieve the status of files in the specified Git repository.

    Parameters
    ----------
    repo_path : str, optional
        The path to the Git repository. Defaults to the current directory.

    Returns
    -------
    dict
        A dictionary with the following keys:

        - 'added': list of str
            List of file paths for files that have been added.
        - 'modified': list of str
            List of file paths for files that have been modified.
        - 'deleted': list of str
            List of file paths for files that have been deleted.
        - 'renamed': list of str
            List of file paths for files that have been renamed.
    """
    # Open the repository
    repo = pygit2.Repository(repo_path)

    # Get the current branch reference
    try:
        current_branch = repo.head
    except pygit2.GitError as e:
        raise pygit2.GitError(
            f"{repo_path} is not a git repo or does not have a head") from e

    # Get the main branch reference
    main_branch = repo.branches["main"]

    # Get the commit objects
    current_commit = repo[current_branch.target]
    main_commit = repo[main_branch.target]

    # Get the diff between the current commit and the main branch commit
    diff = repo.diff(main_commit, current_commit)

    # Prepare dictionaries to store file statuses
    status = {
        "added": [],
        "modified": [],
        "deleted": [],
        "moved": [],
    }

    # Process the diff
    for delta in diff.deltas:
        if delta.status == pygit2.GIT_DELTA_ADDED:
            status["added"].append(delta.new_file.path)
        elif delta.status == pygit2.GIT_DELTA_MODIFIED:
            status["modified"].append(delta.new_file.path)
        elif delta.status == pygit2.GIT_DELTA_DELETED:
            status["deleted"].append(delta.old_file.path)
        elif delta.status == pygit2.GIT_DELTA_RENAMED:
            status["moved"].append((delta.old_file.path, delta.new_file.path))

    return {
        "added": status["added"],
        "modified": status["modified"],
        "deleted": status["deleted"],
        "moved": status["moved"],
    }


def print_out_git_changed_files(print_function, repo_path="."):
    """Print out changes as detected by Git in format Brassy expects in changlogs."""
    status = get_git_status(repo_path=repo_path)
    for entry in status:
        print_function(f"    {entry}:")
        for file in status[entry]:
            print_function(f"    - '{file}'")


def get_current_git_branch(sanitize=True):
    """
    Get the current dirs git branch name.

    Parameters
    ----------
    sanitize : bool, optional
        If True, sanitize branch name as a valid file name before returning.

    Returns
    -------
    str
        The name of the current git branch.
    """
    repo = pygit2.Repository(".")
    branch = repo.head.shorthand
    if sanitize:
        valid_characters = "-_." + string.ascii_letters + string.digits
        return ''.join(c if c in valid_characters else "-" for c in branch)
    else:
        return branch



"""Manages getting and setting settings."""

import os
from pathlib import Path

import platformdirs
import pygit2
import yaml
from pydantic import ValidationError

from brassy.templates.settings_template import SettingsTemplate


def get_git_repo_root(path="."):
    """
    Find the root directory of the Git repository for a path.

    Parameters
    ----------
    path : str, optional
        Path inside the repository. Defaults to ".".

    Returns
    -------
    Path
        Absolute path to the repository root (the dir that contains .git).
    """
    return (Path(pygit2.Repository(path).path) / "..").resolve()


def get_project_config_file_path(app_name):
    """
    Return the path to the project's configuration file.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    Path
        Path to the project's configuration file. If the file does not
        exist locally, the path is resolved relative to the repository
        root when possible.
    """
    project_file = Path(f".{app_name}")
    if project_file.is_file():
        return project_file
    try:
        return get_git_repo_root() / project_file
    except pygit2.GitError:
        return project_file


def get_user_config_file_path(app_name):
    """
    Retrieve the user-specific configuration file path for the app.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    Path
        Path to the user's configuration file.
    """
    return Path(platformdirs.user_config_dir(app_name)) / "user.config"


def get_site_config_file_path(app_name):
    """
    Retrieve the site-wide configuration file path for the app.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    Path
        Path to the site's configuration file.
    """
    return Path(platformdirs.site_config_dir(app_name)) / "site.config"


def get_config_files(app_name):
    """
    Get configuration file paths in increasing precedence.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    List[Path]
        List of configuration file paths. Site, user, then project.
    """
    config_files = []
    for f in [
        get_site_config_file_path,
        get_user_config_file_path,
        get_project_config_file_path,
    ]:
        path = f(app_name)
        config_files.append(path)
    return config_files


def create_config_file(config_file):
    """
    Create a configuration file with default settings.

    Parameters
    ----------
    config_file : Path
        Path where the configuration file will be created.
    """
    config_file = Path(config_file)
    default_settings = SettingsTemplate()
    config_dir = config_file.parent
    if config_dir:
        config_dir.makedirs(exist_ok=True, parents=True)
    with config_file.open("w") as f:
        yaml.dump(default_settings.model_dump(), f)


def read_config_file(config_file, create_file_if_not_exist=False):
    """
    Read and parse a YAML configuration file.

    Parameters
    ----------
    config_file : Path or str
        Path to the configuration file.
    create_file_if_not_exist : bool
        Creates file if it doesn't exist

    Returns
    -------
    dict
        Parsed configuration settings as a dictionary.
    """
    config_file = Path(config_file)
    try:
        with config_file.open() as f:
            return yaml.safe_load(f)
    except FileNotFoundError:
        if not create_file_if_not_exist:
            return SettingsTemplate().model_dump()
        else:
            create_config_file(config_file)
            return read_config_file(config_file)


def merge_and_validate_config_files(config_files):
    """
    Merge settings from multiple config files and validate them.

    Parameters
    ----------
    config_files : list of Path
        Paths to configuration files. Later files override earlier ones.

    Returns
    -------
    dict
        Merged and validated configuration settings.

    Raises
    ------
    ValidationError
        If any file's settings fail to validate against the
        SettingsTemplate model.
    """
    settings = {}
    for config_file in config_files:
        file_settings = read_config_file(config_file, create_file_if_not_exist=False)
        try:
            SettingsTemplate(**file_settings)
        except ValidationError as e:
            print(f"Failed to validate {config_file}")
            print(repr(e.errors()[0]))
            raise
        settings.update(file_settings)
    return settings


def get_settings_from_config_files(app_name):
    """
    Retrieve settings from configuration files without env overrides.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    dict
        Configuration settings merged from files.
    """
    return merge_and_validate_config_files(get_config_files(app_name))


def override_dict_with_environmental_variables(input_dict):
    """Override dict values with case insensitive environment variables when available.

    Parameters
    ----------
    input_dict : dict
        Original settings dictionary.

    Returns
    -------
    dict
        Updated settings dictionary with environment variable overrides.
    """
    env_vars = dict(os.environ)
    lower_env_vars = {
        key.lower(): {"env_var": key, "value": value} for key, value in env_vars.items()
    }
    for key in input_dict:
        if key.lower() in lower_env_vars:
            override = lower_env_vars[key.lower()]
            input_dict[key] = override["value"]
    return input_dict


def get_settings(app_name):
    """
    Return final application settings with file and env overrides.

    Parameters
    ----------
    app_name : str
        Name of the application.

    Returns
    -------
    SettingsTemplate
        An instance containing the merged configuration.

    Raises
    ------
    ValidationError
        If the final settings fail to validate against the model.
    """
    file_settings = override_dict_with_environmental_variables(
        get_settings_from_config_files(app_name),
    )
    settings = SettingsTemplate(**file_settings)
    return settings



"""Base init, sets __version__."""

import logging

try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__package__ or __name__)
except ModuleNotFoundError:
    try:
        import importlib_metadata

        __version__ = importlib_metadata.version(__package__ or __name__)
    except ModuleNotFoundError:
        logging.debug(
            "Could not set __version__ because importlib.metadata is not available."
            + "If using python 3.7, install importlib-metadata to fix this issue",
        )



"""Provides functions for 'pruning' yaml of extra and empty fields."""

from pathlib import Path

import yaml


def prune_empty(data, prune_lists=True, key=""):
    """
    Recursively remove empty values from a nested dictionary or list.

    Parameters
    ----------
    data : dict or list
        The data structure to prune.
    prune_lists : bool, optional
        Indicates whether to prune empty lists. Currently unused.
    key : str, optional
        The key associated with the current data item, used for special cases.

    Returns
    -------
    dict or list or None
        The pruned data structure, or None if it is empty.

    Notes
    -----
    The function considers the following values as empty: `None`, empty strings,
    empty dictionaries, and empty lists. If a value is `0` and the key is
    `"number"`, it is also considered empty to address the related issues
    field which was previously set to 0 instead of null.

    Examples
    --------
    >>> data = {'a': None, 'b': '', 'c': {'d': [], 'e': 'value'}}
    >>> prune_empty(data)
    {'c': {'e': 'value'}}
    """
    nulls = (None, "", {}, [])
    if isinstance(data, dict):
        pruned = {k: prune_empty(v, key=k, prune_lists=prune_lists)
                  for k, v in data.items()}
        pruned = {k: v for k, v in pruned.items() if v not in nulls}
        return pruned if pruned else None
    elif isinstance(data, list) and prune_lists:
        pruned = [prune_empty(item, prune_lists=prune_lists) for item in data]
        pruned = [item for item in pruned if item not in nulls]
        return pruned if pruned else None
    elif data == 0 and key == "number":
        return None
    else:
        return data


def prune_yaml_file(yaml_file_path, console):
    """
    Prune empty values from a YAML file and overwrite it with the pruned content.

    Parameters
    ----------
    yaml_file_path : str
        The file path to the YAML file to be pruned.
    console : Console
        An object used for printing messages to the console.

    Returns
    -------
    None

    Notes
    -----
    This function reads the YAML file, prunes empty values using `prune_empty`,
    and writes the pruned content back to the same file.

    Examples
    --------
    >>> prune_yaml_file('config.yaml', console)
    Pruned config.yaml
    """
    yaml_file_path = Path(yaml_file_path)
    with yaml_file_path.open("r+") as file:
        content = yaml.safe_load(file)
        file.seek(0)
        file.write(
            yaml.dump(
                prune_empty(content, prune_lists=True),
                sort_keys=False,
                default_flow_style=False,
            ),
        )
        file.truncate()
    console.print(f"Pruned {yaml_file_path}")


def direct_pruning_of_files(input_files_or_folders, console, working_dir):
    """
    Prune empty values from YAML files specified by input paths.

    Parameters
    ----------
    input_files_or_folders : list of str
        A list of file paths or directories containing YAML files to prune.
    console : Console
        An object used for printing messages to the console.
    working_dir : str
        The working directory path.

    Returns
    -------
    None

    Notes
    -----
    This function collects YAML files from the specified input paths and
    prunes each file using `prune_yaml_file`.

    Examples
    --------
    >>> direct_pruning_of_files(['configs/'], console, '/home/user')
    Pruned configs/config1.yaml
    Pruned configs/config2.yaml
    """
    import brassy.utils.CLI  # noqa: PLC0415 # here to prevent circular import

    yaml_files = brassy.utils.CLI.get_file_list_from_cli_input(
        input_files_or_folders,
        console,
        working_dir=working_dir,
    )
    for yaml_file in yaml_files:
        prune_yaml_file(yaml_file, console)



"""Wrapper for CLI call and top-level functions."""

from __future__ import annotations

from brassy.utils import settings_manager

Settings = settings_manager.get_settings("brassy")


def run_from_CLI(): # noqa: N802
    """Run brassy via the CLI."""
    from brassy.utils import CLI  # noqa: PLC0415 # here to prevent circular import
    CLI.run_from_CLI()


if __name__ == "__main__":
    run_from_CLI()



"""Handle file system I/O."""

import sys
from pathlib import Path

import yaml
from pygit2 import GitError

from brassy.brassy import Settings
from brassy.templates.release_yaml_template import ReleaseNote
from brassy.utils import git_handler


def get_yaml_template_path(file_path_arg, working_dir=None):
    """
    Return path of the YAML template file based on the given file path argument.

    Args:
        file_path_arg (str): The file path argument provided by the user.

    Returns
    -------
        str: The path of the YAML template file.

    """
    if working_dir is None:
        working_dir = Path.getcwd()
    if file_path_arg is None:
        filename = Path(f"{git_handler.get_current_git_branch()}.yaml")
        return working_dir / filename
    else:
        file_path_arg = Path(file_path_arg)
    if ("/" in str(file_path_arg) or "\\" in str(file_path_arg)
        or Path(file_path_arg).is_file()):
        return file_path_arg
    return working_dir / file_path_arg


def create_blank_template_yaml_file(file_path_arg, console, working_dir="."):
    """
    Create a blank YAML template file with a predefined structure.

    This function generates a YAML file at the specified path with a default
    template. It handles special characters required for YAML compatibility and writes
    the file to disk.

    Parameters
    ----------
    file_path_arg : str
        The file path of the YAML template as passed via the CLI.
    console : rich.console.Console
        A Rich Console object used for displaying messages and errors to the user.
    working_dir : str, optional
        The working directory path. Defaults to the current directory ".".

    Raises
    ------
    SystemExit
        If a Git repo is not found in the current working directory and no file path
        is provided, the program exits with an error message.

    Notes
    -----
    This function performs a string replacement to insert a "|" due to an issue with
    YAML's handling of pipe symbols. For more details, see:
    https://github.com/yaml/pyyaml/pull/822
    """
    pipe_replace_string = "REPLACE_ME_WITH_PIPE"

    default_yaml = {
        category: [
            {
                "title": "",
                "description": (
                    pipe_replace_string
                    if Settings.description_populates_with_pipe
                    else ""
                ),
                "files": {change: [""] for change in Settings.valid_changes},
                "related-issue": {"number": None, "repo_url": ""},
                # in time, extract from the first and last commit
                "date": {"start": None, "finish": None},
            },
        ]
        for category in Settings.change_categories
    }
    try:
        yaml_template_path = get_yaml_template_path(file_path_arg, working_dir)
    except GitError:
        console.print(
            "[bold red]Could not find a git repo. Please run in a "
            + "git repo or pass a file path for the yaml template "
            + "(eg '-t /path/to/file.yaml').",
        )
        sys.exit(1)
    with Path(yaml_template_path).open("w") as file:
        yaml_text = yaml.safe_dump(
            default_yaml,
            sort_keys=False,
            default_flow_style=False,
        )
        if Settings.description_populates_with_pipe:
            yaml_text = yaml_text.replace(pipe_replace_string, "|\n    replace_me")
        file.write(yaml_text)


def value_error_on_invalid_yaml(content, file_path):
    """
    Check if the YAML content follows the correct schema.

    Parameters
    ----------
    content : dict
        Parsed content of the YAML file.
    file_path : str
        Path to the YAML file.

    Raises
    ------
    ValueError
        If the YAML content does not follow the correct schema.
    """
    if content is None:
        raise ValueError(f"No valid brassy-related YAML. Please populate {file_path}")

    ReleaseNote(**content)


def read_yaml_files(input_files, rich_open):
    """
    Read and parse the given list of YAML files.

    Parameters
    ----------
    input_files : list
        List of paths to the YAML files.

    Returns
    -------
    dict
        Parsed content of all YAML files categorized by type of change.

    Examples
    --------
    >>> read_yaml_files(["file1.yaml", "file2.yaml"])
    {'bug-fix': [
        {'title': 'fixed explosions',
          'description': 'This fixed the explosion mechanism'},
        {'title': 'fixed cats not being cute',
          'description': 'This made the cats WAY cuter'}
        ]
    }
    """
    data = {}
    for file_path in input_files:
        with rich_open(file_path, "r", description=f"Reading {file_path}") as file:
            content = yaml.safe_load(file)
            value_error_on_invalid_yaml(content, file_path)
            for category, entries in content.items():
                filtered_entries = [
                    entry
                    for entry in entries
                    if not (entry["title"] == "" or entry["description"] == "")
                ]
                if category not in data and len(filtered_entries) > 0:
                    data[category] = []
                if len(filtered_entries) > 0:
                    data[category].extend(filtered_entries)
    return data


def write_output_file(output_file, content):
    """
    Write the formatted release notes to the output file.

    Parameters
    ----------
    output_file : str
        Path to the output .rst file.
    content : str
        Formatted release notes.
    """
    with Path(output_file).open("w") as file:
        file.write(content)



"""Handle outputs/inputs to the CLI."""

import logging

import rich
import rich.progress
from rich.console import Console as rich_console  # noqa: N813
from rich.logging import RichHandler
from rich.prompt import Confirm
from rich.traceback import install as install_rich_tracebacks

logging.captureWarnings(True)


def init_logger(use_rich):
    """
    Initialize and configure the logger.

    Parameters
    ----------
    use_rich : bool
        If True, sets up rich logging else use standard stream logging

    Returns
    -------
    logger : logging.Logger
        The configured logger instance
    """
    logger = logging.getLogger("build_docs")
    if use_rich:
        install_rich_tracebacks()
        logging_handlers = [RichHandler(rich_tracebacks=True)]
    else:
        logging_handlers = [logging.StreamHandler()]

    logging.basicConfig(level=logging.DEBUG, datefmt="[%X]", handlers=logging_handlers)
    logger.debug("Program initialized")
    return logger


def get_rich_opener(no_format=False):
    """
    Return opener function with or without a rich progress bar.

    Args:
        no_format (bool, optional):
            If True, returns the opener function without any formatting.
            If False, returns the opener function with formatting. Defaults to False.

    Returns
    -------
        function: The opener function for rich progress bar.
    """
    if no_format:
        return rich.progress.Progress().open
    else:
        return rich.progress.open


def setup_console(no_format=False, quiet=False):
    """
    Set up and return the console for printing messages.

    Args:
        no_format (bool, optional): Whether to disable formatting. Defaults to False.
        quiet (bool, optional): Whether to suppress console output. Defaults to False.

    Returns
    -------
        Console: The configured rich console object.
    """
    if not no_format:
        install_rich_tracebacks()
    console = rich_console(quiet=quiet, no_color=(no_format or quiet))
    return console


def get_boolean_prompt_function(enable_format=True):
    """Return a function that prompts Y/N and returns True/False."""
    if enable_format:
        return Confirm.ask
    else:

        def bool_prompt(question):
            answer = input(question).lower()
            if answer in ["yes", "y", "ye"]:
                return True
            elif answer in ["no", "n", ""]:
                return False
            else:
                print("Please respond with 'yes' or 'no'.")
                return bool_prompt(question)

        return bool_prompt


def setup_messages(enable_format, quiet):
    """Set up global objects for outputting messages to the CLI."""
    global open  # noqa: PLW0603
    global boolean_prompt  # noqa: PLW0603
    global RichConsole  # noqa: PLW0603
    global print  # noqa: PLW0603
    open = get_rich_opener(no_format=not enable_format)  # noqa: A001
    RichConsole = setup_console(no_format=not enable_format, quiet=quiet)
    print = RichConsole.print if enable_format else print  # noqa: A001
    boolean_prompt = get_boolean_prompt_function(enable_format=enable_format)


open = get_rich_opener()  # noqa: A001
RichConsole = setup_console()
print = RichConsole.print  # noqa: A001
boolean_prompt = get_boolean_prompt_function()



"""Build release note output."""

import sys
from datetime import datetime

import brassy
from brassy.brassy import Settings
from brassy.utils.messages import RichConsole as console  # noqa: N813


def get_header_footer(rich_open, header_file=None, footer_file=None):
    """Read optional header and footer content.

    Parameters
    ----------
    rich_open : Callable
        Context manager factory for reading files.
    header_file : str or None, optional
        Path to header file.
    footer_file : str or None, optional
        Path to footer file.

    Returns
    -------
    tuple of str or None
        Header and footer content or None when unavailable.
    """

    def get_file(file):
        if not file:
            return None
        with rich_open(file, "r", description=f"Reading {file}") as open_file:
            return open_file.read()

    return get_file(header_file), get_file(footer_file)


def find_duplicate_titles(data):
    """Detect duplicate titles across changelog entries.

    Parameters
    ----------
    data : dict
        Mapping of categories to lists of changelog entries.

    Returns
    -------
    bool
        True if any title occurs more than once.
    """
    titles = [entry["title"] for category in data for entry in data[category]]
    return len(set(titles)) != len(titles)


def format_files_changed_entry(detailed, entry): # noqa: ARG001 TODO: Fix this unused arg
    """Format an RST block describing changed files for an entry.

    Parameters
    ----------
    detailed : bool
        Unused flag kept for compatibility.
    entry : dict
        Changelog entry containing file changes.

    Returns
    -------
    str
        RST formatted file change listing.
    """
    files_changed = "::\n\n"
    for change_type in entry["files"]:
        files_changed += "".join(
            [
                f"    {change_type}: {file}\n"
                for file in filter(lambda x: x != "", entry["files"][change_type])
            ],
        )
    return files_changed


def generate_file_change_section_list_of_strings(
    entry,
    line,
    category,
    title,
    description,
):
    """Create file-specific section lines for a changelog entry.

    Parameters
    ----------
    entry : dict
        Changelog entry with file change data.
    line : str
        Template string for the section line.
    category : str
        Entry category name.
    title : str
        Resolved entry title.
    description : str
        Resolved entry description.

    Returns
    -------
    list of str
        Section lines formatted per file change.
    """
    lines = []
    for change_type in entry["files"]:
        if "{file}" in line:
            for file in filter(lambda x: x != "", entry["files"][change_type]):
                filename = file
                if "{" in filename:
                    filename = filename.replace("{","{{")
                if "}" in filename:
                    filename = filename.replace("}","}}")
                lines.append(
                    line.format(
                        change_type=category.capitalize(),
                        title=title,
                        description=description,
                        file_change=change_type,
                        file=filename,
                    ),
                )
        else:
            lines.append(
                line.format(
                    change_type=category.capitalize(),
                    title=title,
                    description=description,
                    file_change=change_type,
                ),
            )
    return lines


def generate_section_string( # noqa: PLR0913,PLR0912 TODO: Fix complexity of this function
    section_lines,
    changelog_entries,
    release_date,
    version,
    footer,
    header,
):
    """Render a changelog section from templates and entries.

    Parameters
    ----------
    section_lines : list of str
        Template lines for the section.
    changelog_entries : dict
        Mapping of categories to changelog entries.
    release_date : str
        Release date string.
    version : str
        Release version string.
    footer : str or None
        Footer content appended to templates.
    header : str or None
        Header content prepended to templates.

    Returns
    -------
    str
        Rendered section content.
    """
    lines = []
    entry_keywords = [
        "{" + k + "}"
        for k in ["title", "description", "file_change", "file", "change_type"]
    ]
    if any(keyword in line for keyword in entry_keywords for line in section_lines):
        for category, entries in changelog_entries.items():
            for entry in entries:
                if not entry["title"] and not entry["description"]:
                    continue
                if entry["title"]:
                    title = entry["title"]
                    title = title.capitalize()
                else:
                    title = Settings.default_title
                if entry["description"]:
                    description = entry["description"]
                else:
                    description = Settings.default_description
                for line in section_lines:
                    if "{file_change}" in line:
                        lines.extend(
                            generate_file_change_section_list_of_strings(
                                entry,
                                line,
                                category,
                                title,
                                description,
                            ),
                        )
                    else:
                        lines.append(
                            line.format(
                                change_type=category.capitalize(),
                                title=title,
                                description=description,
                            ),
                        )
    else:
        for line in section_lines:
            lines.append(line)
    for i, line in enumerate(lines):
        lines[i] = line.format(
            prefix_file=header,
            suffix_file=footer,
            release_version=version,
            release_date=release_date,
        )
    return "\n".join(lines)


def format_release_notes(data, version, release_date=None, header=None, footer=None):
    """Generate release notes content from parsed changelog data.

    Parameters
    ----------
    data : dict
        Parsed changelog entries grouped by category.
    version : str or None
        Release version string.
    release_date : str or None, optional
        Release date override in ISO format. Defaults to today's date.
    header : str or None, optional
        Optional header content.
    footer : str or None, optional
        Optional footer content.

    Returns
    -------
    str
        Release notes rendered in RST.
    """
    if release_date is None:
        release_date = datetime.now().strftime("%Y-%m-%d")

    header = header or ""
    footer = footer or ""

    release_template = Settings.templates.release_template
    formatted_string = ""
    for section in release_template:
        for _section_name, lines in section.items():
            formatted_string = (
                formatted_string
                + generate_section_string(
                    lines,
                    data,
                    release_date,
                    version,
                    footer,
                    header,
                )
                + "\n"
            )
    if Settings.default_title in formatted_string:
        console.print(
            "Warning: Build completed, but at least one title is missing.",
            style="yellow",
        )
    if Settings.default_description in formatted_string:
        console.print(
            "Warning: Build completed, but at least one description is missing.",
            style="yellow",
        )
    return formatted_string.strip()


def build_release_notes( # noqa PLR0913
    input_files_or_folders,
    console,
    rich_open,
    version=None,
    release_date=None,
    header_file=None,
    footer_file=None,
    working_dir=".",
):
    """Build release notes by reading YAML files and templates.

    Parameters
    ----------
    input_files_or_folders : list of str
        CLI-supplied file or directory paths.
    console : Console
        Console for user feedback.
    rich_open : Callable
        Context manager factory for reading files.
    version : str or None, optional
        Release version string.
    release_date : str or None, optional
        Release date override in ISO format.
    header_file : str or None, optional
        Path to an optional header file.
    footer_file : str or None, optional
        Path to an optional footer file.
    working_dir : str, optional
        Base directory for relative paths.

    Returns
    -------
    str
        Rendered release notes in RST.
    """
    yaml_files = brassy.utils.CLI.get_file_list_from_cli_input(
        input_files_or_folders,
        console,
        working_dir=working_dir,
    )
    try:
        data = brassy.utils.file_handler.read_yaml_files(yaml_files, rich_open)
    except (ValueError, TypeError) as e:
        console.print(f"[red]{e}")
        sys.exit(1)
    header, footer = get_header_footer(
        rich_open,
        header_file=header_file,
        footer_file=footer_file,
    )
    content = format_release_notes(
        data,
        version=version,
        release_date=release_date,
        header=header,
        footer=footer,
    )
    return content



"""Release note YAML validation logic."""

from __future__ import annotations

from datetime import date as Date  # noqa N812
from typing import Dict, List  # noqa: UP035

import dateparser
from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
    RootModel,
    field_validator,
    model_validator,
)

from brassy.utils.settings_manager import get_settings


class InvalidDateValueError(ValueError):
    """Error for invalid date strings."""

    def __init__(self, date_string: str) -> None:
        super().__init__(f"Invalid type for date field: {date_string}")


class Files(BaseModel):
    """Files model for validating files impacted in the changelog."""

    deleted: List[str] = []  # noqa: UP006
    moved: List[str] = []  # noqa: UP006
    added: List[str] = []  # noqa: UP006
    modified: List[str] = []  # noqa: UP006

    @model_validator(mode="after")
    def check_at_least_one_field(self):
        """Verify that at least one file is provided."""
        if not any(
            getattr(self, field) for field in ["deleted", "moved", "added", "modified"]
        ):
            raise ValueError(
                "At least one of deleted, moved, added, or modified must have a value",
            )
        return self


class RelatedInternalIssue(BaseModel):
    """Pydantic class for 'internal' or non-public related issue."""

    internal: str | None = Field(pattern=r"[A-Za-z]+#\d+ - .+", default=None)


class RelatedIssue(BaseModel):
    """Pydantic class for validating related issue (eg. GitHub issue)."""

    number: int | List[int] | None = None  # noqa: UP006
    repo_url: HttpUrl | None = None

    @field_validator("repo_url", mode="before")
    @classmethod
    def convert_empty_to_none(cls, value):
        """Convert empty strings to None for URL validation."""
        if value == "":
            return None
        return value


class DateRange(BaseModel):
    """Date range model for pydantic validation."""

    start: Date | None = None
    finish: Date | None = None

    @field_validator("start", "finish", mode="before")
    @classmethod
    def parse_date(cls, value):
        """
        Parse and validate date values.

        Converts various date formats to a Date object, handling strings,
        Date objects, and None values.

        Args:
            value: Input to parse (Date, None, or string)

        Returns
        -------
            Date or None: Parsed date or None for empty values

        Raises
        ------
            InvalidDateValue: If the value cannot be parsed as a valid date
        """
        # Already correct type or None
        if value is None or isinstance(value, Date):
            return value

        # String values
        if isinstance(value, str):
            value = value.strip()
            if not value or value.lower() in ("never", "null"):
                return None

            parsed = dateparser.parse(
                value,
                settings={
                    "PARSERS": [
                        "timestamp",
                        "relative-time",
                        "absolute-time",
                        "no-spaces-time",
                    ],
                },
            )
            if parsed is None:
                raise InvalidDateValueError(f"Could not parse date: {value}")
            return parsed.date()
        raise InvalidDateValueError(f"Unsupported value type: {type(value)}")

    @model_validator(mode="after")
    def validate_date_range(self):
        """Validate that finish date is not before start date if both are set."""
        if self.start and self.finish and self.finish < self.start:
            raise ValueError("Finish date cannot be before start date")
        return self


class ChangeItem(BaseModel):
    """A model representing a change "item", or an atomic change.

    This class provides a structured way to represent changes with associated metadata
    such as title, description, affected files, related issues, and date range.

    Parameters
    ----------
    title : str or None, optional
        The title of the change item. Must be at least 1 character long if provided.
        Whitespace is stripped.
    description : str or None, optional
        A detailed description of the change.
        Must be at least 1 character long if provided.
        Whitespace is stripped.
    files : Files
        The files affected by this change.
    related_issue : RelatedIssue, RelatedInternalIssue, or None, optional
        An issue related to this change. Aliased as "related-issue" in serialized form.
        Default is None.
    date : DateRange or None, optional
        The date range associated with this change. Default is None.

    Notes
    -----
    Empty strings for 'title' and 'description' are automatically converted to None
    during validation.
    """

    title: str | None = Field(min_length=1, strip_whitespace=True)
    description: str | None = Field(min_length=1, strip_whitespace=True)
    files: Files
    related_issue: RelatedIssue | RelatedInternalIssue | None = Field(
        alias="related-issue",
        exclude_unset=True,
        default=None,
    )
    date: DateRange | None = None

    @model_validator(mode="before")
    def empty_str_to_none(self):
        """
        Convert empty strings to None for 'title' and 'description' attributes.

        This method checks if the 'title' or 'description' attributes of the object
        are empty strings and converts them to None if they are.

        Returns
        -------
        self : object
        Returns the instance itself to allow for method chaining.
        """
        for value in ["title", "description"]:
            if self[value] == "":
                self[value] = None
        return self


class ReleaseNote(RootModel[Dict[str, List[ChangeItem]]]):  # noqa: UP006
    """ReleaseNote is a root model for Release Notes.

    It contains a dictionary that maps category names to lists of ChangeItems.
    """

    pass


Settings = get_settings("brassy")

# List of categories stored in a variable
categories = Settings.change_categories



"""The CLI only portions of brassy.

Brassy can be run without this file, and importing it brassy without it should
allow users to call brassy without the CLI
"""

import argparse
import os
import sys
from pathlib import Path

from rich_argparse import RichHelpFormatter

import brassy.actions.build_release_notes
import brassy.actions.init
import brassy.actions.prune_yaml
import brassy.utils.file_handler
import brassy.utils.git_handler
from brassy.utils import messages
from brassy.utils.settings_manager import get_settings

Settings = get_settings("brassy")


def get_parser():
    """
    Return ArgumentParser for CLI.

    Returns
    -------
        argparse.ArgumentParser: The ArgumentParser object with predefined arguments.
    """
    parser = argparse.ArgumentParser(
        description="Generate release notes from YAML files."
        + " Entries are sorted by order in yaml files, "
        + "and by order of yaml files provided via the command line.",
        formatter_class=RichHelpFormatter,
    )
    parser.add_argument(
        "-t",
        "--write-yaml-template",
        type=str,
        help="Write template YAML to provided file."
        + " If folder provided, place template in folder with current "
        + "git branch name as file name.",
        nargs="?",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-c",
        "--get-changed-files",
        type=str,
        help="Print git tracked file changes against main."
        + " If directory provided, use that directories checked-out branch.",
        nargs="?",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "input_files_or_folders",
        type=str,
        nargs="*",
        help="The folder(s) containing YAML files and/or YAML files. "
        + "Folders will be searched recursively.",
    )
    parser.add_argument(
        "-r",
        "--release-version",
        type=str,
        default="[UNKNOWN]",
        help="Version number of the release. Default is '[UNKNOWN]'.",
        required=False,
        dest="version",
    )
    parser.add_argument(
        "-d",
        "--release-date",
        type=str,
        help="Date of the release. Default is current system time.",
        required=False,
    )
    parser.add_argument(
        "-nc",
        "--no-color",
        action="store_true",
        default=Settings.use_color,
        help="Disable text formatting for CLI output.",
    )
    parser.add_argument(
        "-p",
        "--prefix-file",
        type=str,
        help="A header file to prepend to the release notes.",
    )
    parser.add_argument(
        "-s",
        "--suffix-file",
        type=str,
        help="A footer file to suffix to the release notes.",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        help="The output file for release notes.",
    )
    if Settings.default_yaml_path and Settings.enable_experimental_features:
        yaml_path = "." / Settings.default_yaml_path
    else:
        yaml_path = "."
    parser.add_argument(
        "-yd",
        "--yaml-dir",
        type=str,
        help="Directory to write yaml files to",
        default=yaml_path,
    )
    parser.add_argument(
        "--output-to-console",
        action="store_true",
        default=False,
        help="Write generated release notes to console.",
    )
    parser.add_argument(
        "-nr",
        "--no-rich",
        action="store_true",
        help="Disable rich text output",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Only output errors")
    parser.add_argument(
        "-pr",
        "--prune",
        action="store_true",
        help="Prune provided yaml file(s) of empty sections",
    )
    parser.add_argument(
        "--init",
        action="store_true",
        help="Initialize brassy and generate config files",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="store_true",
        dest="print_version",
        help="Print program version and exit",
        default=False,
    )
    return parser


def parse_arguments():
    """
    Parse command line arguments for input folder and output file.

    Returns
    -------
    argparse.Namespace
        Parsed arguments containing input_folder and output_file.
    """
    parser = get_parser()
    return parser.parse_args(), parser


def print_version_and_exit():
    """Print version and exit with status code 0."""
    messages.RichConsole.print(f"Brassy is at version {brassy.__version__}")
    sys.exit(0)


def exit_on_invalid_arguments(args, parser, console):
    """
    Validate the argparse arguments.

    This function validates the provided argparse arguments to ensure
    that the required input files/folders and output file are provided.
    If arguments are invalid, it prints an error message and exits the program.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed arguments.

    parser : argparse.ArgumentParser
        The ArgumentParser object used to parse the command-line arguments.
    """
    if (
        bool(args.input_files_or_folders)
        or "get_changed_files" in args
        or args.init
        or args.version
    ):
        return

    if "write_yaml_template" in args:
        return

    console.print("[bold red]Invalid arguments.\n")
    parser.print_help()
    sys.exit(1)


def get_yaml_files_from_input(input_files_or_folders):
    """
    Get a list of YAML files from the given input files or folders.

    Parameters
    ----------
    input_files_or_folders : list
        List of paths to input files or folders.

    Returns
    -------
    list
        List of paths to YAML files.

    Raises
    ------
    ValueError
        If a file is not a YAML file or if no YAML files are found in a directory.
    """
    yaml_files = []
    for path in input_files_or_folders:
        if Path(path).is_file():
            if path.suffix not in [".yaml", ".yml"]:
                raise ValueError(f"File {path} is not a YAML file.")
            yaml_files.append(path)
        elif path.is_dir():
            dir_yaml_files = []
            for root, _dirs, files in os.walk(path):
                for file in files:
                    if file.endswith(".yaml") or file.endswith(".yml"):
                        dir_yaml_files.append(Path(root) / Path(file))
            if len(dir_yaml_files) == 0:
                raise FileExistsError(f"No YAML files found in directory {path}.")
            yaml_files.extend(dir_yaml_files)
        else:
            raise FileNotFoundError(path)
    return yaml_files


def get_file_list_from_cli_input(input_files_or_folders, console, working_dir="."):
    """Parse CLI input string into full file paths."""
    try:
        yaml_files = get_yaml_files_from_input(
            [
                brassy.utils.file_handler.get_yaml_template_path(
                    path,
                    working_dir=working_dir,
                )
                for path in input_files_or_folders
            ],
        )
    except FileExistsError as e:
        if Settings.fail_on_empty_dir:
            console.print(f"[red]Invalid file or directory: [bold]{e}[/]")
            sys.exit(1)
        else:
            console.print(f"[yellow]Invalid file or directory: [bold]{e}[/]")
            console.print(
                "[yellow]Returning 0 because fail_on_empty_dir is [bold]False[/]",
            )
            sys.exit(0)
    except FileNotFoundError as e:
        console.print(f"[red]Invalid file or directory: [bold]{e}[/]")
        sys.exit(1)
    except ValueError as e:
        console.print(f"[red]{e}")
        sys.exit(1)
    return yaml_files


def run_from_CLI(): # noqa: N802
    """Generate release notes from YAML files and write output file."""
    args, parser = parse_arguments()

    messages.setup_messages(enable_format=not args.no_rich, quiet=args.quiet)

    console = messages.RichConsole
    printer = messages.print
    rich_open = messages.open

    exit_on_invalid_arguments(args, parser, console)
    if args.print_version:
        print_version_and_exit()
    elif args.init:
        brassy.actions.init.init()
        sys.exit(0)
    elif args.prune:
        brassy.actions.prune_yaml.direct_pruning_of_files(
            args.input_files_or_folders,
            console,
            args.yaml_dir,
        )
    elif "write_yaml_template" in args:
        brassy.utils.file_handler.create_blank_template_yaml_file(
            args.write_yaml_template,
            console,
            working_dir=args.yaml_dir,
        )
    elif "get_changed_files" in args:
        brassy.utils.git_handler.print_out_git_changed_files(
            printer,
            repo_path=args.get_changed_files if args.get_changed_files else ".",
        )
    elif args.input_files_or_folders:
        content = brassy.actions.build_release_notes.build_release_notes(
            args.input_files_or_folders,
            console,
            rich_open,
            version=args.version,
            release_date=args.release_date,
            header_file=args.prefix_file,
            footer_file=args.suffix_file,
            working_dir=args.yaml_dir,
        )
        if args.output_file:
            brassy.utils.file_handler.write_output_file(args.output_file, content)
            if not args.quiet:
                console.print(f"[green]Wrote release notes to {args.output_file}")
        elif not args.quiet:
            console.print(
                "[green]Release notes built successfully. No output file provided.",
            )
        if args.output_to_console:
            console.print(content)

    else:
        parser.print_help()



"""Initialize brassy with config file."""

import brassy
from brassy.brassy import Settings


def init():
    """
    Initialize configuration files for the application.

    This function creates site and user configuration files, and optionally
    a project configuration file based on user input.

    Returns
    -------
    None

    Examples
    --------
    >>> init()
    Do you want to create a project config file? [y/N]: y
    """
    if Settings.enable_experimental_features:
        conf_files = [
            brassy.utils.settings_manager.get_site_config_file_path("brassy"),
            brassy.utils.settings_manager.get_user_config_file_path("brassy"),
        ]
    else:
        conf_files = []
    for conf_file in conf_files:
        brassy.utils.settings_manager.create_config_file(conf_file)
    if brassy.utils.messages.boolean_prompt(
        "Do you want to create a project config file?",
    ):
        brassy.utils.settings_manager.create_config_file(
            brassy.utils.settings_manager.get_project_config_file_path("brassy"),
        )
